/*=======================================================================*
 |   file name : heightmapper.cpp
 |-----------------------------------------------------------------------*
 |   function  :
 |-----------------------------------------------------------------------*
 |   author    :
 *=======================================================================*/

/**************************** i n c l u d e s ****************************/

#include "platform.h"
#include "switches.h"
#include "echidna\ensure.h"

#include "echidna\argparse.h"
#include "echidna\eerrors.h"
#include "echidna\readgfx.h"
#include "echidna\eio.h"
#include "echidna\checkglu.h"

#include <stdio.h>
#include <math.h>

/*************************** c o n s t a n t s ***************************/


/******************************* t y p e s *******************************/


/************************** p r o t o t y p e s **************************/


/***************************** g l o b a l s *****************************/


/****************************** m a c r o s ******************************/


/**************************** r o u t i n e s ****************************/

struct vec3
{
    vec3(float x, float y, float z) : x(x), y(y), z(z) { }
	float normalize();

	float x;
	float y;
	float z;
};

#define eps		          float(10e-6)

float vec3::normalize()
{
	float norm = sqrtf(x * x + y * y + z * z);
	if (norm > eps)
		norm = 1.0f / norm;
	else
		norm = 0.0f;
	x *= norm;
	y *= norm;
	z *= norm;
	return norm;
}

/*************************************************************************
                              saveTGA32Bit
 *************************************************************************

   SYNOPSIS
		int saveTGA32Bit (int fh, BlockO32BitPixels *bop)

   PURPOSE


   INPUT
		fh  :
		bop :

   OUTPUT
		None

   EFFECTS
		None

   RETURNS


   SEE ALSO


   TODO
 	* Don't use CHK functions cause they exit


   HISTORY
		07/15/96 : Created.

 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

typedef struct TGAHeader
{
	uint8	ID;				// Byte 0
	uint8	ctype;			// Byte 1
	uint8	itype;			// Byte 2
	uint8	mincolorl;		// Byte 3
	uint8	mincolorh;		// Byte 4
	uint8	colorsl;		// Byte 5
	uint8	colorsh;		// Byte 6
	uint8	colorsize;		// Byte 7
	uint8	originxl;		// Byte 8
	uint8	originxh;		// Byte 9
	uint8	originyl;		// Byte 10
	uint8	originyh;		// Byte 11
	uint8	widthl;			// Byte 12
	uint8	widthh;			// Byte 13
	uint8	heightl;		// Byte 14
	uint8	heighth;		// Byte 15
	uint8	bpp;			// Byte 16
	uint8	idesc;			// Byte 17
} TGAHeader;


int saveTGA8Bit (int fh, BlockO8BitPixels *bop)
{
	static TGAHeader tgaHeader =
	{
	0,	//	uint8	ID;				// Byte 0
	0,	//	uint8	ctype;			// Byte 1
	3,	//	uint8	itype;			// Byte 2 -- uncompressed 8bit greyscale
	0,	//	uint8	mincolorl;		// Byte 3
	0,	//	uint8	mincolorh;		// Byte 4
	0,	//	uint8	colorsl;		// Byte 5
	0,	//	uint8	colorsh;		// Byte 6
	0,	//	uint8	colorsize;		// Byte 7
	0,	//	uint8	originxl;		// Byte 8
	0,	//	uint8	originxh;		// Byte 9
	0,	//	uint8	originyl;		// Byte 10
	0,	//	uint8	originyh;		// Byte 11
	0,	//	uint8	widthl;			// Byte 12
	0,	//	uint8	widthh;			// Byte 13
	0,	//	uint8	heightl;		// Byte 14
	0,	//	uint8	heighth;		// Byte 15
	8,	//	uint8	bpp;			// Byte 16
	0,	//	uint8	idesc;			// Byte 17
	};

	uint8 *buffer;

	buffer = (uint8 *)malloc (bop->width * bop->height);
	if (!buffer)
	{
		SetGlobalErr (ERR_GENERIC);
		GEcatf ("Out of memory saving tga");
		return FALSE;
	}

	tgaHeader.widthl = (UINT8)(bop->width % 256);
	tgaHeader.widthh = (UINT8)(bop->width / 256);

	tgaHeader.heightl = (UINT8)(bop->height % 256);
	tgaHeader.heighth = (UINT8)(bop->height / 256);

	CHK_Write (fh, &tgaHeader, sizeof (tgaHeader));

	memcpy (buffer, bop->pixels, bop->width * bop->height);
	flipBuffer (buffer, bop->width * 1, bop->height);
	CHK_Write (fh, buffer, bop->width * bop->height * 1);
	free (buffer);

	return TRUE;

}

/************************** a r g u m e n t s  ***************************/

#define	ARG_INFILE		(newargs[ 0])
#define ARG_OUTFILE		(newargs[ 1])

char Usage[] =	"Usage: normalmipper INFILE OUTFILE\n"
				"    converts an normalmap to a mip of 1/2 the width and height\n"
;

ArgSpec Template[] = {
{STANDARD_ARG|REQUIRED_ARG,				"INFILE",	"\tINFILE   = in targa/photoshop\n", },
{STANDARD_ARG|REQUIRED_ARG,				"OUTFILE",	"\tOUTFILE  = targa\n", },
{0, NULL, NULL, },
};

int main(int argc, char** argv)
{
	char			**newargs;

	newargs = argparse (argc, argv, Template);

	if (!newargs)
	{
		EL_printf ("%s\n", GlobalErrMsg);
		printarghelp (Usage, Template);
		RETURN EXIT_FAILURE;
	}
	else
	{
		BlockO32BitPixels* pBOP;

		pBOP = Read32BitPicture (ARG_INFILE);
		if (!pBOP)
		{
			printf ("Error: %s (%s)\n", GlobalErrMsg, ARG_INFILE);
			exit (1);
		}

		if (((pBOP->width & 1) != 0) || ((pBOP->height & 1) != 0))
		{
			printf ("Error: width and height must be even (%s)\n", ARG_INFILE);
			exit (1);
		}

		{
			BlockO32BitPixels mip;

			mip.width    = pBOP->width / 2;
			mip.height   = pBOP->height / 2;
			mip.channels = 0;

			long size = mip.width * mip.height;

			pixel32* pSrc = pBOP->rgba;
			pixel32* pDst = (pixel32*)CHK_AllocateMemory (size * sizeof(pixel32), "dest pixel buffer");

			mip.rgba = pDst;

			for (int yy = 0; yy < mip.height; ++yy)
			{
				for (int xx = 0; xx < mip.width; ++xx)
				{
					vec3 norm(0.0f, 0.0f, 0.0f);

					for (int sy = 0; sy < 2; ++sy)
					{
						for (int sx = 0; sx < 2; ++sx)
						{
							pixel32* p = pSrc + (yy * 2 + sy) * pBOP->width + xx * 2 + sx;

							norm.x += (float)(p->red   / 255.0f) * 2.0f - 1.0f;
							norm.y += (float)(p->green / 255.0f) * 2.0f - 1.0f;
							norm.z += (float)(p->blue  / 255.0f) * 2.0f - 1.0f;
						}
					}

					norm.normalize();

					pixel32* d = pDst + (yy) * mip.width + xx;

					d->red   = (uint8)((norm.x + 1.0f) * 255.0 / 2.0f);
					d->green = (uint8)((norm.y + 1.0f) * 255.0 / 2.0f);
					d->blue  = (uint8)((norm.z + 1.0f) * 255.0 / 2.0f);
					d->alpha = 255;
				}
			}

			if (!Write32BitPicture (ARG_OUTFILE, &mip))
			{
				printf ("Error: %s (%s)\n", GlobalErrMsg, ARG_OUTFILE);
				exit (1);
			}
		}
	}
	return 0;
}

